  #include "wfh.h"
  #include "wordtree.h"
  #include <stdlib.h> /* malloc/free */
  #include <assert.h>
  
  struct wfh_struct {
    WORDTREE* tree;
  };
  
  WFH* wfh_create(wfh_compare comp)
  {
    WFH* pwfh;

  /* Precondition:                              */
  /*   comp != 0                                */
    assert(comp != NULL);

    pwfh = (WFH*)malloc(sizeof(WFH));
    if (pwfh != NULL)
    {
      pwfh->tree = wordtree_create(comp);
      if (pwfh->tree == NULL)
      {
        free(pwfh);
        pwfh = 0;
      }
    }
    return pwfh;
  }

  unsigned wfh_insert(WFH* hist,
                      const char* word)
  {
     unsigned** ppcount;
  #ifndef NDEBUG
  /* for post condition: */
    unsigned old_freq;
  #endif

  /* Preconditions:                             */
  /*   hist != NULL                             */
    assert(hist != NULL);

  /*   word != NULL                             */
    assert(word != NULL);

  
  #ifndef NDEBUG
  /* for post condition: */
    old_freq = wfh_find(hist, word);
  #endif
    
    ppcount = (unsigned**)wordtree_insert(hist->tree,
					  word);
    if (ppcount == NULL)
    {
      return 0; /* insertion failed */
    }
    if (*ppcount == NULL) /* first time seen */
    {
      *ppcount = (unsigned*)malloc(sizeof(unsigned));
      **ppcount = 1;
    }
    else
    {
      ++(**ppcount);
    }
  /* Postconditions:                            */
  /* wfh_find(hist,word) ==                      */
  /*   1+old wfh_find(hist,word)                 */
    assert(wfh_find(hist, word) == 1+old_freq);
    return **ppcount;
  }


  unsigned wfh_find(WFH* hist, const char* word)
  {
    unsigned** pfreq;
  /* Preconditions:                             */
  /*   hist != NULL                             */
    assert(hist != NULL);
  /*   word != NULL                             */

    assert(word != NULL);

    /* wordtree_find returns the pointer to  */
    /* userdata pointer. If the pointer      */
    /* is NULL, the word isn't in the tree.  */
    /* If it's non-NULL, we can interpret it */
    /* as a pointer to an unsigned int       */
    /* since that's how we allocated it when */
    /* inserting the word) and get the       */
    /* frequency from there.                 */

    pfreq = (unsigned**)wordtree_find(hist->tree,
                                     word);

    return pfreq == NULL ? 0 : **pfreq;
  }

  struct travdata {
    wfh_visitor visitor;
    void* usercommon;
  };

  static int treevisitor(const char* word,
                         void* data,
			 void* commondata)
  {
    struct travdata* ptd = (struct travdata*)commondata;

    /* get the visotor the user intended to call */
    wfh_visitor userfunc = ptd->visitor;

    unsigned count = *(unsigned*)data;

    /* Call the user's visitor with the word,    */
    /* the frequency and the user's commondata.  */
    return userfunc(word, count, ptd->usercommon);
  }

  int wfh_traverse(WFH* hist,
		   wfh_visitor visitor,
		   void* commondata)
  {
    struct travdata ourdata;

    /* Preconditions:                           */
    /*   hist != NULL                           */
    assert(hist != NULL);

    /*   visitor != NULL                        */
    assert(visitor != NULL);

    /* store the user supplied visitor and      */
    /* commondata in ourdata.                   */
    ourdata.visitor = visitor;
    ourdata.usercommon = commondata;

    /* Traverse the tree with our tree visitor  */
    /* and pass it the user supplied visitor    */
    /* and common data through ourdata passed   */
    /* as the commondata. treevisitor knows how */
    /* to unpack this and forward the call to   */
    /* the user specified visitor function.     */
    return wordtree_traverse(hist->tree,
                             treevisitor,
			     &ourdata);
  }

  void wfh_destroy(WFH* hist)
  {
  /* Preconditions:                             */
  /*   hist != NULL                             */
    assert(hist != NULL);

    wordtree_destroy(hist->tree, free);
    free(hist);
  }
