#include <stdio.h>        /* size_t and file stuff */
#include "wordfile.h"
#include <assert.h>       /* assert                */
#include <stdlib.h>       /* malloc and free       */
#include <ctype.h>        /* is***                 */
#include <string.h>       /* strchr                */

struct wordfile_struct { /* our ADT */
      FILE* file;
      wordfunction isWord;
      void* userdata;
      int firstCharInWord;
};

/* the default wordfunction, which uses "isalnum" */

static charkind wordfile_isword(unsigned char c, void* userdata) 
{                                          /* static to avoid polluting */
   return isalnum(c) ? inWord : notInWord;  /* the global namespace      */
   /* return "inWord" if "isalnum" returns non-0, and notInWord otherwise */
}

/* the string testing function */
static charkind wordfile_string(unsigned char c, void* userdata)
{
   return (c != 0 && strchr((const char*)userdata, c) != NULL)    /* 1 */
      ? inWord : notInWord;         /* same construction as above */
}


WORDFILE* wordfile_open(const char* filename)
{
   /* First create the wordfile to use */
   WORDFILE* wordfile = (WORDFILE*)malloc(sizeof(WORDFILE));

    /* check that malloc succeeded */
   if (wordfile == NULL)
   {
      return NULL;
   }

   /* Preconditions:                                        */
   /*   filename != NULL                                    */
   assert(filename != NULL);

    /* Now open the file, and assign the "file" component    */
    /* of our newly created wordfile the value returned by   */
    /* fopen                                                 */
   wordfile->file = fopen(filename, "r");

   if (wordfile->file == NULL)
   {                  
      free(wordfile);  /* if fopen failed, free the allocated memory */
      wordfile = NULL; /* and set the return value to NULL           */
   }
   else
   {
      wordfile->firstCharInWord = EOF;       /* No first char to use,    */
      wordfile->userdata = NULL;           /* and no user data set     */
      wordfile->isWord = &wordfile_isword; /* More initializations, we */
      /* want "isalnum" default   */
      /* behaviour.               */
   }
   return wordfile;
}


/* close is exactly the same as before */

int wordfile_close(WORDFILE* wordfile)
{
   int retval = 1;
   if (wordfile != NULL)
   {
      retval = (fclose(wordfile->file) == 0);
      free(wordfile);
   }
   return retval;
}

size_t wordfile_nextword(WORDFILE* wordfile,
			 char* buffer,
			 size_t buffersize)
{
   size_t retval = 0;
   int c = EOF;
   charkind kind; /* this variable is used to control the logic      */
   /* when determining whether to store the character */
   /* for the next iteration. `kind' should only be   */
   /* accessed if `c' is not EOF. The logic is fairly */
   /* convoluted to guarantee exactly one call of     */
   /* "wordfile->isWord" for every character.         */


   /* Preconditions:                                               */
   /*   wordfile != NULL.                                          */
   assert(wordfile != NULL);

   /*   buffersize >= 2                                            */
   assert(buffersize >= 2);

   /*   buffer != NULL                                             */
   assert(buffer != NULL);

   /* internal check. If wordfile->file == NULL, something is      */
   /* seriously wrong.                                             */
   assert(wordfile->file != NULL);

   /* obtain the first character of the word. If we have a "leftover" */
   /* first character from the last call, pretend that we've just     */
   /* read it from the file                                           */

   if (wordfile->firstCharInWord != EOF)
   {
      /* restore state from end of previous call */
      c = wordfile->firstCharInWord;
      kind = firstInWord;
      wordfile->firstCharInWord = EOF; /* don't handle again in next */
      /* call, unless specifically  */
      /* set otherwise below        */
   }
   else /* otherwise we must take care of non-word characters */
   {
      do
      {
	 c = fgetc(wordfile->file);
	 if (c != EOF)
	    kind = wordfile->isWord(c, wordfile->userdata);  /* 2 */
      } while (c != EOF && kind == notInWord);
      /* loop until we find EOF or something in a word */
   }

   /* no matter which way we took above, "c" is now either EOF or a    */
   /* character that belongs to a word. The following loop is not      */
   /* terminated by a firstInWord in the first iteration.              */

   while (c != EOF && (kind == inWord
		       || (kind == firstInWord && retval == 0)))
   {
      if (retval < buffersize)
	 buffer[retval++] = (char)c;
      /* read the next character and keep `kind' up to date */
      c = fgetc(wordfile->file);
      if (c != EOF)
	 kind = wordfile->isWord((unsigned char)c, wordfile->userdata);
   }

   /* now we've either hit EOF, a character of type notInWord, or  */
   /* a character of type firstInWord. If the character that broke */
   /* the loop above was firstInWord, it must be stored for the    */
   /* next call                                                    */

   if (c != EOF && kind == firstInWord)
      wordfile->firstCharInWord = c; /* store it for use in next call */

   if (retval < buffersize)
      buffer[retval] = 0;/* null terminate */
   else
      buffer[buffersize-1] = 0; /* force null-termination of too long word */

   return retval; /* return the length of the copied word */
}



void wordfile_wordfunction(WORDFILE* wordfile,
			   wordfunction function,
			   void* userdata)
{
   /* Preconditions:                                                     */
   /*   wordfile != NULL                                                 */

   assert(wordfile != NULL);

   /*   Current word definition is the default definition                */

   assert(wordfile->isWord == &wordfile_isword);

   /*   No words are read from the word file                             */

   assert(ftell(wordfile->file) == 0L);                     /* 4 */

   wordfile->isWord = function;
   wordfile->userdata = userdata;
}


void wordfile_wordchars(WORDFILE* wordfile, const char* string)
{
   /* Preconditions:                                                     */
   /*   wordfile != NULL                                                 */

   assert(wordfile != NULL);

   /*   Current word definition is the default definition                */

   assert(wordfile->isWord == &wordfile_isword);

   /*   No words are read from the word file                             */

   assert(ftell(wordfile->file) == 0L);                     /* 4 */
    
   wordfile->isWord = &wordfile_string;
   wordfile->userdata = (void*)string;
}
