  #include <stdlib.h> /* malloc/free */
  #include <assert.h> /* assertions */
  #include <string.h> /* strlen, strcpy */
  #include "wordtree.h" /* the prototypes */

  typedef struct treenode_struct {
    struct treenode_struct* pLeft;
    struct treenode_struct* pRight;
    char* word;
    void* userdata;
  } TREENODE;

  struct wordtree_struct {
    TREENODE* pRoot;
    wordtree_comparison comp;
  };

  static TREENODE** findnode(TREENODE** node,
                             wordtree_comparison comp,
                             const char* word);

  /* Helper function for wordtree_insert and         */
  /* wordtree_find. Returns the address of a pointer */
  /* to a node. If a node with the right word is     */
  /* found, the address of the pointer to that node  */
  /* is returned. If no node is found, the address   */
  /* of the pointer to be initialised with the new   */
  /* node is returned.                               */


  static void subtreedestroy(TREENODE* node,
                             wordtree_userdatadel del);
  /* Helper function for destruction. It traverses   */
  /* tree and calls del (if not NULL) for the        */
  /* userdata of every node, before deallocating the */
  /* node itself.                                    */


  static int subtreetraverse(TREENODE* node,
                             wordtree_wordvisitor visitor,
			     void* always);
  /* Helper function for wordtree_traverse.          */
  /* Traverses the entire tree from node and         */
  /* downwards and calls visitor in sorted order     */


  WORDTREE* wordtree_create(wordtree_comparison comp)
  {
    WORDTREE* result=(WORDTREE*)malloc(sizeof(WORDTREE));
    if (result != NULL)
    {
      result->pRoot = NULL; /* initialise empty tree */
      result->comp = comp;  /* with comparison func. */
    }
    return result;
  }


  void** wordtree_insert(WORDTREE* tree, 
                         const char* word)
  {
    TREENODE** theNode;

    /* Preconditions:                              */
    /*   tree != NULL                              */
    assert(tree != NULL);

    /*   word != NULL                              */
    assert(word != NULL);

    /* Get pointer to pointer to the node with our */
    /* word. If the word is not in the tree, the   */
    /* pointer points to a null pointer, which we  */
    /* must initialise. Otherwise we can use the   */
    /* returned value directly.                    */
    theNode = findnode(&(tree->pRoot),
                       tree->comp,
                       word);

    if (*theNode == NULL) /* word is not yet in */
    {                     /* tree insert it now */
      *theNode = (TREENODE*)malloc(sizeof(TREENODE));
      (*theNode)->pLeft = NULL;
      (*theNode)->pRight = NULL;
      (*theNode)->userdata = NULL;
      (*theNode)->word = (char*)malloc(strlen(word)+1);
      strcpy((*theNode)->word, word);
    }
    
    return &((*theNode)->userdata);
  }
 

  void** wordtree_find(WORDTREE* tree,
                      const char* word)
  {
    TREENODE** theNode;
 
    /* Preconditions:                              */
    /*   tree != NULL                              */
    assert(tree != NULL);

    /*   word != NULL                              */
    assert(word != NULL);

    /* Just as with insert above, we get a pointer */
    /* to a pointer to the node. If the pointer    */
    /* pointed to is NULL, the word isn't there.   */
    theNode = findnode(&(tree->pRoot),
                       tree->comp,
                       word);

    return (*theNode == NULL) ? NULL
                              : &((*theNode)->userdata);
  }


  int wordtree_traverse(WORDTREE* tree,
                        wordtree_wordvisitor visitor,
			void* always)
  {    
    /* Preconditions:                               */
    /*   tree != NULL                               */
    assert(tree != NULL);

    /*   visitor != NULL                            */
    assert(visitor != NULL);

    return subtreetraverse(tree->pRoot, visitor, always);
  }


  void wordtree_destroy(WORDTREE* tree,
                        wordtree_userdatadel deletor)
  {
    /* Preconditions:                         */
    /*   tree != NULL                         */
    assert(tree != NULL);

    /* Destroy all nodes in the tree, and the */
    /* associated user data (if applicable)   */
    subtreedestroy(tree->pRoot, deletor);

    /* Then deallocate the wordtree struct    */
    free(tree);
  }

  /********************/
  /* Helper functions */
  /*******************/

  static TREENODE** findnode(TREENODE** node,
                             wordtree_comparison comp,
                             const char* word)
  { /* See description in beginning of file */
    TREENODE** npp = node;
    while (*npp != NULL)
    {
      /* compare word, with the word in *node */
      int result = comp(word, (*npp)->word);

      if (result == 0) /* word == (*npp)->word */
	{
	  return npp; /* OK, got it, return it. */
	}
      if (result < 0) /* word < (*npp)->word go on */
      {               /* searching to the left.   */
	  npp = &((*npp)->pLeft); /* Addr. of pointer */
	}
      else /* word > (*npp)->word, go on  */
      {    /* searching to the right     */
	npp = &((*npp)->pRight); /* Address of pointer */
      }
    }
    return npp;    
  }

  static void subtreedestroy(TREENODE* node,
                             wordtree_userdatadel del)
  { /* See description in beginning of file, and */
    /* detailed description of tree traversal    */
    /* further down.                             */
    if (node != NULL)
    {
      /* First destroy everything to the left */
      subtreedestroy(node->pLeft, del);

      /* Then everything to the right */
      subtreedestroy(node->pRight, del);

      /* Then the user data, if del allows us */
      if (del != NULL)
	{
	  del(node->userdata);
	}

      /* And then deallocate this node */
      free(node);
    }
  }

  static int subtreetraverse(TREENODE* node,
                             wordtree_wordvisitor visitor,
			     void* commondata)
  { /* See description in beginning of file, and */
    /* detailed description of tree traversal    */
    /* further down.                             */
    if (node == NULL)
    {
      return 1;
    }
    return subtreetraverse(node->pLeft,
			   visitor,
			   commondata)
        && visitor(node->word,
		   node->userdata,
		   commondata)
        && subtreetraverse(node->pRight,
			   visitor,
			   commondata);

    /* Here we make use of C's lazy evaluation of */
    /* logical expressions.                       */
    /*                                            */
    /* If the first call to subtreetraverse       */
    /* returns 0, the whole expression will be    */
    /* zero, so nothing more needs to be done. In */
    /* this case visitor and the second           */
    /* subtreetraverse will not be called.        */
    /* The same happens if visitor returns 0,     */
    /* then the whole expression will be zero     */
    /* anyway, so the last subtreetraverse will   */
    /* not be called.                             */
    /*                                            */
    /* In other words, if any of the above        */
    /* functions returns 0, the rest of them will */
    /* not be called (thus aborting the           */
    /* traversal) and the function will return 0. */
    /* This zero return value will then be        */
    /* propagated upwards, immediately terminating*/
    /* all traversals.                            */
  }
